# Automatically generated by scripts/generate_extension_properties.py
# **DO NOT MODIFY - CHANGES WILL BE OVERWRITTEN**

from .base import ExtensibleClass


class LDAPExtensions(ExtensibleClass):

    @property
    def netgroups(self):
        """The built-in NIS netgroups extension

        :rtype: laurelin.extensions.netgroups.LaurelinLDAPExtension
        """
        return self._get_extension_instance('netgroups')
